/*
**    ltspinfod
**
**    Linux Terminal Server Printer Daemon
**    Version 0.1
**    James McQuillan (jam@ltsp.org), October 2003
**    Placed under GPL.
**
**    This daemon will run on a workstation, listening on a port (9200)
**    for a command to run.  Initially, the first command will be a 'get'
**    command, for retrieving configuration things, like the
**    sound daemon entry.
*/

#include    <unistd.h>
#include    <stdio.h>
#include    <string.h>
#include    <stdlib.h>
#include    <fcntl.h>
#include    <netdb.h>
#include    <syslog.h>
#include    <errno.h>
#include    <ctype.h>
#include    <sys/types.h>
#include    <sys/resource.h>
#include    <sys/stat.h>
#include    <sys/socket.h>
#include    <netinet/in.h>
#include    <arpa/inet.h>
#include    <popt.h>
#include    <dirent.h>
#include    <sys/wait.h>

#define     FALSE           0
#define     TRUE            1
#define     INFODIR         "/tmp/info"
#define     MAXARGS         20

#define     LOGOPTS         (LOG_PERROR|LOG_PID|LOG_DAEMON|LOG_ERR)

static char version[] = "ltspinfod Version 0.1";

static int  nPort          = 9200;
static int  fDebug         = FALSE;
static int  fAllowShutdown = FALSE;
static int  fAllowProcRead = FALSE;
static int  fNoDaemon      = FALSE;

char delims[] = " \t\n";                      // SPACE, TAB, NEWLINE 

void DecodeCommandLine(int, char **);
int  split( char *, char **);
void process_getcfg( char **, int, int );
void process_getproc( char **, int, int );
void process_shutdown( );
void process_reboot( );
void get_value( char *, char * );

//-----------------------------------------------------------------------------

int command_loop( int fd ){
  int  nread;
  unsigned char buffer[1024];
  char *args[MAXARGS];

  if(fDebug){
    fprintf(stderr,"about to read the command\n");
  }

  nread = read(fd, buffer, sizeof(buffer)-1 );

  if( nread > 0 ){

    int argc = 0;

    buffer[nread] = '\0';		// Terminate the buffer

    argc = split( buffer, args );

    if(fDebug){
      int i;
      printf("argc = %d\n", argc);
      for( i=0; i < argc; i++ ){
        printf("args[%d] = [%s]\n", i, args[i]);
      }
    }

    if( argc > 0 ){

      //
      // Dispatching is done here, based on the first
      // element of the array
      //

      if( strcasecmp( args[0], "getcfg" ) == 0 ){
        process_getcfg( args+1, argc-1, fd );
      }
      else if( strcasecmp(args[0], "getproc" ) == 0 && fAllowProcRead ){
        process_getproc( args+1, argc-1, fd );
      }
      else if( strcasecmp(args[0], "shutdown" ) == 0 && fAllowShutdown ){
        process_shutdown( );
      }
      else if( strcasecmp(args[0], "reboot" ) == 0 && fAllowShutdown ){
        process_reboot( );
      }
    }
  }
  return(nread);
}

//-----------------------------------------------------------------------------

void process_getcfg( char **arr, int argc, int fd ){

  char filename[128];
  char value[256];
  char outbuf[256];
  char *p;
  int  len;
  int  i;

  for( i = 0; i < argc; i++ ){

    // Convert string to uppercase
    p = arr[i];
    while(*p){
      *p = toupper(*p);
      p++;
    }

    if(strcmp(arr[i],"ALL") == 0){
      DIR *dp;
      struct dirent *de;
      if( (dp = opendir(INFODIR)) ){
        while( (de = readdir(dp)) ){
          struct stat fbuf;
          sprintf(filename,"%s/%s", INFODIR, de->d_name);
          //
          // Look to see what kind of directory entry it is.
          // We are only interested in regular files
          //
          stat( filename, &fbuf );
          if(fDebug){
            printf("fbuf.st_mode=%o\n", fbuf.st_mode);
          }
          if( S_ISREG(fbuf.st_mode) ){
            get_value( filename, value );
            if(fDebug){
              printf("value=[%s]\n",value);
            }
            len = sprintf(outbuf,"%s=\"%s\"\n",de->d_name,value);
            write(fd,outbuf,len);
          }
        }
        closedir(dp);
      }
    }
    else{
      sprintf(filename,"%s/%s", INFODIR, arr[i]);

      get_value( filename, value );

      if(fDebug){
        printf("value=[%s]\n",value);
      }

      len = sprintf(outbuf,"%s=\"%s\"\n",arr[i],value);
      write(fd,outbuf,len);
    }
  }
}

//-----------------------------------------------------------------------------

void get_value( char *filename, char *value ){
  char buf[256];
  FILE *f;
  char *p;

  value[0] = '\0';

  if( ( f = fopen(filename, "r") ) != 0 ){
    fgets(buf,sizeof(buf),f);
    if(strlen(buf) > 0){
      strcpy(value,buf);
    }
    fclose(f);
  }

  // Remove newline from the answer

  p = value;
  while(*p){
    if( *p == '\n' ){
      *p = '\0';
    }
    p++;
  }
}

//-----------------------------------------------------------------------------
//
// process_getproc will read files from the /proc filesystem
//
void process_getproc( char **arr, int argc, int fd ){

  char filename[128];
  char buf[256];
  char outbuf[256];
  FILE *f;
  int  len;
  int  i;

  //
  // We are doing a fork() call, so that we can do a chroot to the
  // /proc directory.  This will prevent evil people from reading other
  // files from the filesystem.
  //
  // When the child process terminates, the parent continues on, waiting
  // for the next request to come in.
  //

  switch( fork() ){
    case -1: syslog(LOGOPTS, "process_getproc: fork: %m\n");
             exit(1);
    case 0:  /* child */
             break;
    default: wait(0);  // Wait for the child to terminate
             return;
  }

  chroot("/proc");

  for( i = 0; i < argc; i++ ){
    sprintf(filename,"/%s", arr[i]);
    if( ( f = fopen(filename, "r") ) != 0 ){
      while(fgets(buf,sizeof(buf),f)){
        len = sprintf(outbuf,"%s",buf);
        write(fd,outbuf,len);
      }
      fclose(f);
    }
    else{
      if(fDebug){
        printf("Error opening %s, errno=%d\n", filename, errno );
      }
    }
  }
  exit(0);   // This is the child exiting
}

//-----------------------------------------------------------------------------
//
// process_shutdown will shutdown the thin client
//
void process_shutdown( ){

  system("/sbin/init 0");

}

//-----------------------------------------------------------------------------
//
// process_reboot will reboot the thin client
//
void process_reboot( ){

  system("/sbin/init 6");

}

//-----------------------------------------------------------------------------

int split( char *str, char **arr ){

  char *p;
  int  argc = 0;

  p = str;

  while( *p ){
    if( (*p == '\n') || (*p == '\r') ){
      *p = '\0';
    }
    p++;
  }

  p = str;

  while( *p && argc < MAXARGS ){

    int len = strspn( p, delims );	// Advance past any leading whitespace
    p += len;

    len = strcspn( p, delims );
    if( len > 0 ){
      arr[argc++] = p;
      p += len;
      *p++ = '\0';
    }
  }

  return(argc);
}

//-----------------------------------------------------------------------------

void server()
{
  struct rlimit       resourcelimit;
  struct protoent     *proto;
  struct sockaddr_in  netaddr, client;

  int    netfd, fd, clientlen, one = 1;
    
  if(fDebug){
    fprintf(stderr,"Inside of server()\n");
  }
  else if( !fNoDaemon ){
    /*
    ** Throw the routine into background by calling fork()
    */

    switch (fork()) {
      case -1: syslog(LOGOPTS, "fork: %m\n");
               exit(1);

      case 0:  /* child */
               break;

      default: /* parent */
               exit(0);
    }

    /* Now in child process */

    /*
    ** Get the maximum number of files that can be opened
    ** by this process so we can close them all.
    */
    resourcelimit.rlim_max = 0;
    if( getrlimit(RLIMIT_NOFILE, &resourcelimit) < 0) {
      syslog(LOGOPTS, "getrlimit: %m\n");
      exit(1);
    }

    /*
    ** Loop through and close all file handles
    */
    for (fd = 0; fd < resourcelimit.rlim_max; ++fd)
      (void)close(fd);

    if( setsid() < 0 ){
      syslog( LOGOPTS, "setsid: %m\n" );
      exit(1);
    }

    (void)chdir("/");
    (void)umask(022);
    fd = open("/dev/null", O_RDWR);    /* stdin */
    (void)dup(fd);                     /* stdout */
    (void)dup(fd);                     /* stderr */

  } /* ! fDebug */



  if( ( proto = getprotobyname("tcp") ) == NULL ){
    syslog(LOGOPTS, "Cannot find protocol for TCP!\n");
    if(fDebug)
      fprintf( stderr, "Error calling getprotobyname() %s\n",
               strerror(errno));
    exit(1);
  }

  if( ( netfd = socket(AF_INET, SOCK_STREAM, proto->p_proto) ) < 0 ) {
    syslog(LOGOPTS, "socket: %m\n");
    if(fDebug)
      fprintf(stderr,"Error calling socket() %s\n",strerror(errno));
    exit(1);
  }

  if( setsockopt(netfd, SOL_SOCKET, SO_REUSEADDR, &one, sizeof(one) ) < 0 ){
    syslog(LOGOPTS, "setsockopt: %m\n");
    if(fDebug)
      fprintf(stderr,"Error calling setsockopt() %s\n",strerror(errno));
    exit(1);
  }

  memset(&netaddr, 0, sizeof(netaddr));
  netaddr.sin_family = AF_INET;
  netaddr.sin_port = htons(nPort);
  netaddr.sin_addr.s_addr = htonl(INADDR_ANY);

  if( bind( netfd, (struct sockaddr *)&netaddr, sizeof(netaddr) ) < 0 ){
    syslog(LOGOPTS, "bind: %m\n");
    if(fDebug)
      fprintf(stderr,"Error calling bind() %s\n",strerror(errno));
    exit(1);
  }

  if( listen(netfd, 5) < 0 ){
    syslog(LOGOPTS, "listen: %m\n");
    if(fDebug)
      fprintf(stderr,"Error calling listen() %s\n",strerror(errno));
    exit(1);
  }

  clientlen = sizeof(client);
  memset(&client, 0, sizeof(client));

  if(fDebug){
    fprintf(stderr,"About to call accept()\n");
  }

  while(( fd = accept( netfd, (struct sockaddr *)&client, &clientlen ) ) >= 0){

    syslog(LOGOPTS, "Connection from %s port %hd\n",
                    inet_ntoa(client.sin_addr),
                    ntohs(client.sin_port));

    if(fDebug){
      fprintf(stderr, "Connection from %s port %hd\n",
                      inet_ntoa(client.sin_addr),
                      ntohs(client.sin_port));
    }

    command_loop( fd );

    if( fDebug ){
      fprintf(stderr,"Command processing complete, going back to loop\n");
    }
    (void)close(fd);
  }

  syslog(LOGOPTS, "accept: %m\n");

  exit(1);
}

//-----------------------------------------------------------------------------

int main(int argc, char **argv)
{
    DecodeCommandLine(argc,argv);

    server();

    closelog();

    return (0);
}

//-----------------------------------------------------------------------------

void usage()
{
  fprintf(stderr,"\n"
                 "Usage: ltspinfod [{-p|--port} <portnum>]    default=9200\n"
		 "                 [{-h|--help}]              This help\n"
		 "                 [{-x|--debug}]             Turn on debugging\n"
		 "                 [{-n|--nodaemon}]          Don't daemonize\n"
                 "                 [{-v|--version}]           Display version\n"
		 "                 [{-s|--allow_shutdown}]    Allow shutdown\n"
		 "                 [{-r|--allow_procread}]    Allow read /proc\n"
    );
  exit(1);
}

//-----------------------------------------------------------------------------

void DecodeCommandLine(int argc, char **argv)
{
  char c;
  poptContext optCon;        // context for parsing command-line options

  struct poptOption optionsTable[] = {
    { "port",           'p', POPT_ARG_INT,    &nPort,        0,   NULL, NULL },
    { "help",           'h', 0,               0,             'h', NULL, NULL },
    { "debug",          'x', 0,               0,             'x', NULL, NULL },
    { "nodaemon",       'n', 0,               0,             'n', NULL, NULL },
    { "version",        'v', 0,               0,             'v', NULL, NULL },
    { "allow_shutdown", 's', 0,               0,             's', NULL, NULL },
    { "allow_procread", 'r', 0,               0,             'r', NULL, NULL },
    { NULL,              0,  0,               NULL,          0,   NULL, NULL }
  };

  optCon = poptGetContext( NULL, argc, (const char **)argv, optionsTable, 0 );

  while( ( c = poptGetNextOpt(optCon) ) >= 0 ){
    if(c < 0){
      fprintf(stderr,"Error occurred!\n");
    }
    switch(c){
      case 'h':   usage();
                  break;
      case 'x':   fDebug = TRUE;
                  break;
      case 's':   fAllowShutdown = TRUE;
                  break;
      case 'r':   fAllowProcRead = TRUE;
                  break;
      case 'n':   fNoDaemon = TRUE;
                  break;
      case 'v':   fprintf(stderr,"%s\n",version);
                  exit(1);
    }
  }

  if( c < -1 ){
    fprintf( stderr,
             "%s, option = %s\n",
             poptStrerror(c),
             poptBadOption(optCon,0) );
    exit(1);
  }

  poptFreeContext(optCon);
}

